"""
Opens links in a browser.

Installation
------------

First use case: local use
~~~~~~~~~~~~~~~~~~~~~~~~~
If you use poezio on your workstation, this is for you.
You only have to load the plugin: ::

    /load link

Second use case: remote use
~~~~~~~~~~~~~~~~~~~~~~~~~~~
If you use poezio through SSH, this is for you.

.. note:: Small explanation: Poezio will create a `Unix FIFO`_ and send the commands in,
            and you will have to run a dæmon locally with ssh, to get those commands.

First, set the :term:`exec_remote` option in the config file to ``true``. Then select
the directory you want to put the fifo in (default is the current
directory, :file:`./`), the :file:`poezio.fifo` file will be created there.

After that, load the plugin: ::

    /load link

And open a link with :term:`/link` (as described below), this will create the FIFO.

You need to grab poezio’s sources on your client computer, or at least the `daemon.py`_
file.

Finally, on your client computer, run the ssh command:

.. code-block:: bash

    ssh toto@example.org "cat ~/poezio/poezio.fifo" | python3 daemon.py

Usage
-----

.. glossary::

    /link
        **Usage:** ``/link [range] [command]``

        This plugin adds a :term:`/link` command that will open the links in
        ``firefox``. If you want to use another browser, or any other
        command, you can use the :term:`/set` command to change the
        :term:`browser` option. You can also specify the command to execute
        directly in the arguments.  For example `/link "mpv %s"` will open
        the first link found using mpv, instead of the configured browser.


        :term:`/link` without argument will open the last link found
        in the current tab, if any is found. An optional range
        argument can be given, to select one or more links to
        open. Examples:
        ``/link 1`` is equivalent to ``/link``
        ``/link 3`` will open the third link found in the current tab,
        starting from the bottom.
        ``/link 1:5`` will open the last five links in the current tab
        ``/link :2`` will open the last two links

Options
-------

:term:`exec_remote`

    To execute the command on your client

.. glossary::

    browser
        Set the default browser started by the plugin

.. _Unix FIFO: https://en.wikipedia.org/wiki/Named_pipe
.. _daemon.py: https://lab.louiz.org/poezio/poezio/raw/main/poezio/daemon.py

"""
import platform
import re

from poezio.plugin import BasePlugin
from poezio.xhtml import clean_text
from poezio import common
from poezio import tabs

url_pattern: re.Pattern[str] = re.compile(
    r'\b'
    r'(?:http[s]?://(?:\S+))|'
    r'(?:magnet:\?(?:\S+))|'
    r'(?:aesgcm://(?:\S+))|'
    r'(?:gopher://(?:\S+))|'
    r'(?:gemini://(?:\S+))'
    r'\b',
    re.I | re.U
)

app_mapping = {
    'Linux': 'xdg-open',
    'Darwin': 'open',
}


class Plugin(BasePlugin):
    def init(self) -> None:
        for _class in (tabs.MucTab, tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                _class,
                'link',
                self.command_link,
                usage='[num] [command]',
                help=(
                    'Opens the last link from the conversation into a browser.\n'
                    'If [num] is given, then it will open the num-th link displayed. '
                    'Use a [command] argument to override the configured browser value.'
                ),
                short='Open links into a browser')

    def find_link(self, nb: int) -> str | None:
        messages = self.api.get_conversation_messages()
        if not messages:
            return None
        for message in messages[::-1]:
            matches = url_pattern.findall(clean_text(message.txt))
            if matches:
                url: str
                for url in matches[::-1]:
                    if nb == 1:
                        return url
                    else:
                        nb -= 1
        return None

    def command_link(self, str_args: str) -> None:
        args = common.shell_split(str_args)
        start = 1
        end = 1
        # With two arguments, the first is the range, the second is the command
        # With only one argument, it is a range if it starts with a number
        # or :, otherwise it is a command
        if len(args) == 2 or\
           len(args) == 1 and (args[0][0].isnumeric() or args[0][0] == ":"):
            if args[0].find(':') == -1:
                try:
                    start = int(args[0])
                    end = start
                except ValueError:
                    self.api.run_command('/help link')
                    return
            else:
                start_str, end_str = args[0].split(':', 1)
                if start_str == '':
                    start_str = "1"
                try:
                    start = int(start_str)
                    end = int(end_str)
                except ValueError:
                    self.api.information('Invalid range: %s' % (args[0]), 'Error')
                    return
        command = None
        if len(args) == 2:
            command = args[1]
        if len(args) == 1 and (not args[0][0].isnumeric()
                               and args[0][0] != ":"):
            command = args[0]
        for nb in range(start, end + 1):
            link = self.find_link(nb)
            if not link:
                self.api.information('No URL found.', 'Warning')
                return
            default = app_mapping.get(platform.system(), 'firefox')
            if command is None:
                self.core.exec_command(
                    [self.config.get('browser', default), link])
            else:
                self.core.exec_command([command, link])

    def cleanup(self) -> None:
        del self.config



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    Utilities
"""

from random import choice

VOWELS = 'aiueo'
CONSONANTS = 'bcdfghjklmnpqrstvwxz'


def pronounceable(length: int = 6) -> str:
    """Generates a pronounceable name"""
    out = ''
    vowels = choice((True, False))
    for _ in range(0, length):
        out += choice(VOWELS if vowels else CONSONANTS)
        vowels = not vowels
    return out



"""
Plugin to kill the TCP transport and force a reconnect.
Useful on bad connections when you want to discard a broken link and
connect a gain. Best used with ``enable_smacks`` set to true.

Usage
-----

.. glossary::

    /tcp-reconnect

        Close the TCP connection abruptly to force a reconnect.
"""

from poezio.plugin import BasePlugin
from poezio.decorators import command_args_parser


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'tcp-reconnect',
            self.reconnect,
            help='Trigger a reconnection by killing the TCP link.',
        )

    @command_args_parser.ignored
    def reconnect(self) -> None:
        if self.core.xmpp.transport:
            self.core.xmpp.transport.abort()
        else:
            self.core.xmpp.reconnect()



"""
Test the functions in the `logger` module
"""
import datetime
from pathlib import Path
from random import sample
from shutil import rmtree
from string import hexdigits
from poezio import logger
from poezio.logger import (
    LogMessage, parse_log_line, parse_log_lines, build_log_message
)
from poezio.ui.types import Message
from poezio.common import get_utc_time
import pytest


class ConfigShim:
    def __init__(self, value):
        self.value = value

    def get_by_tabname(self, name, *args, **kwargs):
        return self.value


logger.config = ConfigShim(True)


@pytest.fixture
def log_dir():
    name = 'tmplog-' + ''.join(sample(hexdigits, 16))
    path = Path('/tmp', name)
    try:
        yield path
    finally:
        rmtree(path, ignore_errors=True)


def read_file(logger, name):
    if '/' in name:
        name = name.replace('/', '\\')
    filename = logger.log_dir / f'{name}'
    with open(filename) as fd:
        return fd.read()


def test_log_roster(log_dir):
    instance = logger.Logger()
    instance.log_dir = log_dir
    instance.log_roster_change('toto@example.com', 'test test')
    content = read_file(instance, 'roster.log')
    assert content[:3] == 'MI '
    assert content[-32:] == ' 000 toto@example.com test test\n'


def test_log_message(log_dir):
    instance = logger.Logger()
    instance.log_dir = log_dir
    msg = Message('content', 'toto')
    instance.log_message('toto@example.com', msg)
    content = read_file(instance, 'toto@example.com')
    line = parse_log_lines(content.split('\n'), '')[0]
    assert line['nickname'] == 'toto'
    assert line['txt'] == 'content'
    msg2 = Message('content\ncontent2', 'titi')
    instance.log_message('toto@example.com', msg2)
    content = read_file(instance, 'toto@example.com')
    lines = parse_log_lines(content.split('\n'), '')

    assert lines[0]['nickname'] == 'toto'
    assert lines[0]['txt'] == 'content'
    assert lines[1]['nickname'] == 'titi'
    assert lines[1]['txt'] == 'content\ncontent2'


def test_parse_message():
    line = 'MR 20170909T09:09:09Z 000 <nick>  body'
    parsed_line = parse_log_line(line, 'user@domain')
    message = LogMessage('2017', '09', '09', '09', '09', '09', '0', 'nick', 'body')

    for slot in LogMessage.__slots__:
        assert getattr(parsed_line, slot) == getattr(message, slot)

    line = '<>'
    assert parse_log_line(line, 'user@domain') is None

    line = 'MR 20170908T07:05:04Z 003 <nick>  '
    parsed_line = parse_log_line(line, 'user@domain') 
    message = LogMessage('2017', '09', '08', '07', '05', '04', '003', 'nick', '')
    for slot in LogMessage.__slots__:
        assert getattr(parsed_line, slot) == getattr(message, slot)


def test_log_and_parse_messages():
    msg1 = {
        'nick': 'toto',
        'msg': 'coucou',
        'date': datetime.datetime.now().replace(microsecond=0),
        'prefix': 'MR',
    }
    msg1_utc = get_utc_time(msg1['date'])
    built_msg1 = build_log_message(**msg1)
    assert built_msg1 == 'MR %s 000 <toto>  coucou\n' % (msg1_utc.strftime('%Y%m%dT%H:%M:%SZ'))

    msg2 = {
        'nick': 'toto',
        'msg': 'coucou\ncoucou',
        'date': datetime.datetime.now().replace(microsecond=0),
        'prefix': 'MR',
    }
    built_msg2 = build_log_message(**msg2)
    msg2_utc = get_utc_time(msg2['date'])
    assert built_msg2 == 'MR %s 001 <toto>  coucou\n coucou\n' % (msg2_utc.strftime('%Y%m%dT%H:%M:%SZ'))

    assert parse_log_lines((built_msg1 + built_msg2).split('\n'), 'user@domain') == [
            {'time': msg1['date'], 'history': True, 'txt': 'coucou', 'nickname': 'toto', 'type': 'message'},
        {'time': msg2['date'], 'history': True, 'txt': 'coucou\ncoucou', 'nickname': 'toto', 'type': 'message'},
    ]



"""
This plugin lets you set the CSI_ state manually, when the autoaway plugin
is not sufficient for your usage.

Commands
--------

.. glossary::

    /csi_active
        **Usage:** ``/csi_active``

        Set CSI state to ``active``.

    /csi_inactive
        **Usage:** ``/csi_inactive``

        Set CSI state to ``inactive``.

.. _CSI: https://xmpp.org/extensions/xep-0352.html
"""

from poezio.plugin import BasePlugin


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'csi_active',
            self.command_active,
            help='Set the client state indication to “active”',
            short='Manual set active')
        self.api.add_command(
            'csi_inactive',
            self.command_inactive,
            help='Set the client state indication to “inactive”',
            short='Manual set inactive')

    def command_active(self, args: str) -> None:
        if not self.core.xmpp.plugin['xep_0352'].enabled:
            self.api.information('CSI is not enabled in this server',
                                 'Warning')
        else:
            self.core.xmpp.plugin['xep_0352'].send_active()

    def command_inactive(self, args: str) -> None:
        if not self.core.xmpp.plugin['xep_0352'].enabled:
            self.api.information('CSI is not enabled in this server',
                                 'Warning')
        else:
            self.core.xmpp.plugin['xep_0352'].send_inactive()



"""
Defines the data-forms Tab
"""

import logging
from typing import Any, Callable, List, TYPE_CHECKING

from slixmpp.exceptions import IqError, IqTimeout

from poezio import windows
from poezio.bookmarks import Bookmark, BookmarkList
from poezio.core.structs import Command
from poezio.tabs import Tab, NoInputTab

from slixmpp import JID, InvalidJID

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class BookmarksTab(NoInputTab):
    """
    A tab displaying lines of bookmarks, each bookmark having
    a 4 widgets to set the jid/password/autojoin/storage method
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', bookmarks: BookmarkList) -> None:
        Tab.__init__(self, core)
        self._name = "Bookmarks"
        self.bookmarks = bookmarks
        self.new_bookmarks: List[Bookmark] = []
        self.removed_bookmarks: List[Bookmark] = []
        self.header_win = windows.ColumnHeaderWin(
            ['name', 'room@server/nickname', 'password', 'autojoin',
             'storage'])
        self.bookmarks_win = windows.BookmarksWin(self.bookmarks)
        self.help_win = windows.HelpText('Ctrl+Y: save, Ctrl+G: cancel, '
                                         '↑↓: change lines, tab: change '
                                         'column, M-a: add bookmark, C-k'
                                         ': delete bookmark')
        self.info_header = windows.BookmarksInfoWin()
        self.key_func['KEY_UP'] = self.bookmarks_win.go_to_previous_line_input
        self.key_func['KEY_DOWN'] = self.bookmarks_win.go_to_next_line_input
        self.key_func['^I'] = self.bookmarks_win.go_to_next_horizontal_input
        self.key_func['^G'] = self.on_cancel
        self.key_func['^Y'] = self.on_save
        self.key_func['M-a'] = self.add_bookmark
        self.key_func['^K'] = self.del_bookmark
        self.resize()
        self.update_commands()

    def add_bookmark(self) -> None:
        new_bookmark = Bookmark(
            JID('room@example.tld/nick'), method='local')
        self.new_bookmarks.append(new_bookmark)
        self.bookmarks_win.add_bookmark(new_bookmark)

    def del_bookmark(self) -> None:
        current = self.bookmarks_win.del_current_bookmark()
        if current in self.new_bookmarks:
            self.new_bookmarks.remove(current)
        elif current:
            self.removed_bookmarks.append(current)

    def on_cancel(self) -> bool:
        self.core.close_tab(self)
        return True

    def on_scroll_down(self) -> bool:
        return self.bookmarks_win.go_to_next_page()

    def on_scroll_up(self) -> bool:
        return self.bookmarks_win.go_to_previous_page()

    def on_save(self) -> None:
        self.bookmarks_win.save()
        if find_duplicates(self.new_bookmarks):
            self.information(
                'Duplicate bookmarks in list (saving aborted)', 'Error')
            return
        for bm in self.new_bookmarks:
            try:
                JID(bm.jid)
                if not self.bookmarks[bm.jid]:
                    self.bookmarks.append(bm)
            except InvalidJID:
                self.information(
                    'Invalid JID for bookmark: %s/%s' % (bm.jid, bm.nick),
                    'Error')
                return

        for bm in self.removed_bookmarks:
            if bm in self.bookmarks:
                self.bookmarks.remove(bm)

        self.core.register_task(
            self.save_routine()
        )

    async def save_routine(self) -> bool:
        try:
            await self.bookmarks.save(self.core.xmpp)
            self.information('Bookmarks saved', 'Info')
        except (IqError, IqTimeout):
            self.information('Remote bookmarks not saved.', 'Error')
        self.core.close_tab(self)
        return True

    def on_input(self, key: str, raw: bool = False) -> None:
        if key in self.key_func:
            res = self.key_func[key]()
            if res:
                return
            self.bookmarks_win.refresh_current_input()
        else:
            self.bookmarks_win.on_input(key, raw=raw)

    def resize(self) -> None:
        self.need_resize = False
        self.header_win.resize_columns({
            'name':
            self.width // 4,
            'room@server/nickname':
            self.width // 4,
            'password':
            self.width // 6,
            'autojoin':
            self.width // 6,
            'storage':
            self.width // 6
        })
        info_height = self.core.information_win_size
        tab_height = Tab.tab_win_height()
        self.header_win.resize(1, self.width, 0, 0)
        self.bookmarks_win.resize(self.height - 3 - tab_height - info_height,
                                  self.width, 1, 0)
        self.help_win.resize(1, self.width, self.height - 1, 0)
        self.info_header.resize(1, self.width,
                                self.height - 2 - tab_height - info_height, 0)

    def on_info_win_size_changed(self) -> None:
        if self.core.information_win_size >= self.height - 3:
            return
        info_height = self.core.information_win_size
        tab_height = Tab.tab_win_height()
        self.bookmarks_win.resize(self.height - 3 - tab_height - info_height,
                                  self.width, 1, 0)
        self.info_header.resize(1, self.width,
                                self.height - 2 - tab_height - info_height, 0)

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        self.header_win.refresh()
        self.refresh_tab_win()
        self.help_win.refresh()
        self.info_header.refresh(self.bookmarks.preferred)
        self.info_win.refresh()
        self.bookmarks_win.refresh()


def find_duplicates(bm_list: list[Bookmark]) -> bool:
    jids = set()
    for bookmark in bm_list:
        if bookmark.jid in jids:
            return True
        jids.add(bookmark.jid)
    return False



from poezio.ui.funcs import (
    find_first_format_char,
    parse_attrs,
    truncate_nick,
)


def test_find_char_not_present():
    assert find_first_format_char("toto") == -1


def test_find_char():
    assert find_first_format_char('a \x1A 1') == 2


def test_truncate_nick():
    assert truncate_nick("toto") == "toto"


def test_truncate_nick_wrong_size():
    assert truncate_nick("toto", -10) == "t…"


def test_truncate_nick_too_long():
    nick = "012345678901234567"
    assert truncate_nick(nick) == nick[:10] + "…"


def test_truncate_nick_no_nick():
    assert truncate_nick('') == ''


def test_parse_attrs():
    text = "\x19o\x19u\x19b\x19i\x191}\x19o\x194}"
    assert parse_attrs(text) == ['4}']


def test_parse_attrs_broken_char():
    text = "coucou\x19"
    assert parse_attrs(text) == []


def test_parse_attrs_previous():
    text = "coucou"
    previous = ['u']
    assert parse_attrs(text, previous=previous) == previous



#!/usr/bin/env python3
import os
import subprocess
import sys
from tempfile import TemporaryFile

try:
    from setuptools import setup, Extension
except ImportError:
    print('\nSetuptools was not found. Install setuptools for python 3.\n')
    sys.exit(1)

cmdclass = {}
current_dir = os.path.dirname(__file__)


def get_relative_dir(folder, stopper):
    """
    Find the path from a directory to a pseudo-root in order to recreate
    the filetree.
    """
    acc = []
    last = os.path.basename(folder)
    while last != stopper:
        acc.append(last)
        folder = os.path.dirname(folder)
        last = os.path.basename(folder)
    return os.path.join(*acc[::-1]) if acc else ''


def find_doc(before, path):
    to_strip = os.path.dirname(os.path.abspath(__file__)) + "/"
    _files = []
    stop = os.path.basename(path)
    for root, dirs, files in os.walk(os.path.join(current_dir, 'doc', path)):
        files_path = []
        relative_root = get_relative_dir(root, stop)
        for name in files:
            files_path.append(os.path.join(root, name).removeprefix(to_strip))
        _files.append((os.path.join(before, relative_root), files_path))
    return _files


def check_include(library_name, header):
    command = [os.environ.get('PKG_CONFIG', 'pkg-config'), '--cflags', library_name]
    try:
        cflags = subprocess.check_output(command).decode('utf-8').split()
    except FileNotFoundError:
        print('pkg-config not found.')
        return False
    except subprocess.CalledProcessError:
        # pkg-config already prints the missing libraries on stderr.
        return False
    command = [os.environ.get('CC', 'cc')] + cflags + ['-E', '-']
    with TemporaryFile('w+') as c_file:
        c_file.write('#include <%s>' % header)
        c_file.seek(0)
        try:
            return subprocess.call(command, stdin=c_file, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
        except FileNotFoundError:
            print('%s headers not found.' % library_name)
            return False


def sphinx_man():
    """
    Requires sphinx-build -b man doc/source doc/build/man
    """
    expected_sphinx_files = [
        'doc/build/man/poezio.cfg.7',
        'doc/build/man/poezio.keys.7',
        'doc/build/man/poezio.commands.7'
    ]
    found = []
    for item in expected_sphinx_files:
        if os.path.exists(item):
            found.append(item)
    if found:
        return [('share/man/man7/', found)]
    return []


sphinx_files_found = sphinx_man()
if not sphinx_files_found:
    print(
        '\nSphinx-built manpages not found. Only the '
        'short handwritten manpages will be installed.\n'
    )


if not check_include('python3', 'Python.h'):
    sys.exit(2)

module_poopt = Extension(
    'poezio.poopt',
    extra_compile_args=['-Wno-declaration-after-statement'],
    sources=['poezio/pooptmodule.c'],
)

# Create a link to the config file (for packaging purposes)
if not os.path.exists(os.path.join(current_dir, 'poezio', 'default_config.cfg')):
    os.link(os.path.join(current_dir, 'data', 'default_config.cfg'),
            os.path.join(current_dir, 'poezio', 'default_config.cfg'))

with open('README.rst', encoding='utf-8') as readme_fd:
    LONG_DESCRIPTION = readme_fd.read()

setup(
    long_description=LONG_DESCRIPTION,
    ext_modules=[module_poopt],
    packages=[
        'poezio', 'poezio.core', 'poezio.tabs', 'poezio.windows',
        'poezio.ui', 'poezio_plugins', 'poezio_themes'
    ],
    package_dir={
        'poezio': 'poezio',
        'poezio_plugins': 'plugins',
        'poezio_themes': 'data/themes'
    },
    package_data={'poezio': ['default_config.cfg', 'py.typed']},
    scripts=['scripts/poezio_logs'],
    entry_points={'console_scripts': ['poezio = poezio.__main__:run']},
    data_files=([
            ('share/man/man1/', ['data/poezio.1', 'data/poezio_logs.1']),
            ('share/poezio/', ['README.rst', 'COPYING', 'CHANGELOG']),
            ('share/applications/', ['data/io.poez.Poezio.desktop']),
            ('share/metainfo/', ['data/io.poez.Poezio.appdata.xml'])
        ]
        + find_doc('share/doc/poezio/source', 'source')
        # Requires sphinx-build doc/source doc/build/html
        + find_doc('share/doc/poezio/html', 'build/html')
        + sphinx_files_found
    ),
    cmdclass=cmdclass,
    command_options={
        'build_man': {
            'builder': ('setup.py', 'man'),
        }
    },
)

# Remove the link afterwards
if (os.path.exists(os.path.join(current_dir, 'poezio', 'default_config.cfg')) and
        os.path.exists(os.path.join(current_dir, 'data', 'default_config.cfg'))):

    os.unlink(os.path.join(current_dir, 'poezio', 'default_config.cfg'))



from __future__ import annotations

import curses

from datetime import (
    datetime,
    date,
)
from functools import singledispatch
from math import ceil, log10
from typing import (
    Any,
    TYPE_CHECKING,
    Sequence,
)

from poezio import poopt
from poezio.theming import (
    get_theme,
    dump_tuple,
    TextStyle,
)
from poezio.ui.consts import (
    FORMAT_CHAR,
)
from poezio.ui.funcs import (
    truncate_nick,
    parse_attrs,
)
from poezio.ui.types import (
    BaseMessage,
    Message,
    StatusMessage,
    UIMessage,
    XMLLog,
)

if TYPE_CHECKING:
    from poezio.windows import Win


# msg is a reference to the corresponding Message object. text_start and
# text_end are the position delimiting the text in this line.
class Line:
    """Generic text line"""
    __slots__ = ('msg', 'start_pos', 'end_pos', 'prepend')

    def __init__(self, msg: BaseMessage, start_pos: int, end_pos: int, prepend: str) -> None:
        self.msg = msg
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.prepend = prepend

    def __repr__(self) -> str:
        return '(%s, %s)' % (self.start_pos, self.end_pos)


class ReactionLine(Line):
    """A line containing only reactions"""
    __slots__ = ('txt')

    def __init__(self, msg: Message, txt: str) -> None:
        super().__init__(msg, start_pos=0, end_pos=0, prepend='')
        self.txt = txt


LinePos = tuple[int, int]


def generate_lines(lines: list[LinePos], msg: BaseMessage,
                   default_color: str = '') -> list[Line]:
    """Generate a number of lines from a message text."""
    line_objects = []
    attrs: list[str] = []
    prepend = default_color if default_color else ''
    for line in lines:
        saved = Line(
            msg=msg,
            start_pos=line[0],
            end_pos=line[1],
            prepend=prepend)
        attrs = parse_attrs(msg.txt[line[0]:line[1]], attrs)
        if attrs:
            prepend = FORMAT_CHAR + FORMAT_CHAR.join(attrs)
        else:
            if default_color:
                prepend = default_color
            else:
                prepend = ''
        line_objects.append(saved)
    return line_objects


@singledispatch
def build_lines(msg: BaseMessage, width: int, timestamp: bool, nick_size: int = 10,
                reactions: bool = True) -> Sequence[Line | None]:
    """Default build_lines dispatcher. Builds a list of lines from a message."""
    offset = msg.compute_offset(timestamp, nick_size)
    lines = poopt.cut_text(msg.txt, width - offset - 1)
    return generate_lines(lines, msg, default_color='')


@build_lines.register(type(None))
def build_separator(*args: Any, **kwargs: Any) -> Sequence[Line | None]:
    """Return a None element for the separator"""
    return [None]


def build_reactions(msg: Message, reactions: dict[str, set[str]], width: int,
                    offset: int) -> list[ReactionLine]:
    """
    Build a list of reactions (displayed in the form "<char>[number]",
    accounting for line length to avoid breaking up the reaction items.
    """
    built: list[list[str]] = [[]]
    current_built = 0
    theme = get_theme()
    info = dump_tuple(theme.COLOR_REACTIONS_TEXT)

    for reaction, ids in sorted(
            reactions.items(),
            key=lambda tup: (len(tup[1]), tup[0]), reverse=True
    ):
        len_ids = len(ids)
        if len_ids > 1:
            post = f'\x19{info}}}[{len_ids}]\x19o'
        else:
            post = ''

        built_reaction = f'{reaction}{post}'
        if current_built + poopt.wcswidth(built_reaction) + 1 > width - offset - 1:
            built.append([])
            current_built = poopt.wcswidth(built_reaction) + 1
        else:
            current_built += poopt.wcswidth(built_reaction) + 1
        built[-1].append(built_reaction)
    prepend = f'\x19{info}}}| '
    return [ReactionLine(msg, prepend + (' '.join(line))) for line in built]


@build_lines.register(Message)
def build_message(msg: Message, width: int, timestamp: bool,
                  nick_size: int = 10, reactions: bool = True) -> Sequence[Line | None]:
    """
    Build a list of lines from this message.
    """
    txt = msg.txt
    if not txt:
        return []
    offset = msg.compute_offset(timestamp, nick_size)
    lines = poopt.cut_text(txt, width - offset - 1)
    generated_lines = generate_lines(lines, msg, default_color='')
    if reactions and msg.reactions:
        built_reactions = build_reactions(msg, msg.reactions, width, offset)
        generated_lines.extend(built_reactions)
    return generated_lines


@build_lines.register(StatusMessage)
def build_status(msg: StatusMessage, width: int, timestamp: bool,
                 nick_size: int = 10, reactions: bool = True) -> Sequence[Line | None]:
    """Build a status message line list."""
    msg.rebuild()
    offset = msg.compute_offset(timestamp, nick_size)
    lines = poopt.cut_text(msg.txt, width - offset - 1)
    return generate_lines(lines, msg, default_color='')


@build_lines.register(XMLLog)
def build_xmllog(msg: XMLLog, width: int, timestamp: bool,
                 nick_size: int = 10, reactions: bool = True) -> Sequence[Line | None]:
    """Build a list of lines for the XML log specific to XML Tab"""
    offset = msg.compute_offset(timestamp, nick_size)
    lines = poopt.cut_text(msg.txt, width - offset - 1)
    return generate_lines(lines, msg, default_color='')


@singledispatch
def write_pre(msg: BaseMessage, win: Win, with_timestamps: bool,
              nick_size: int) -> int:
    """Write the part before text (only the timestamp)"""
    if with_timestamps:
        return PreMessageHelpers.write_time(win, False, msg.time)
    return 0


@write_pre.register(UIMessage)
def write_pre_uimessage(msg: UIMessage, win: Win, with_timestamps: bool, nick_size: int) -> int:
    """ Write the prefix of a ui message log
        - timestamp (short or long)
        - level
    """
    color: TextStyle | None
    offset = 0
    if with_timestamps:
        offset += PreMessageHelpers.write_time(win, False, msg.time)

    if not msg.level:  # not a message, nothing to do afterwards
        return offset

    level = truncate_nick(msg.level, nick_size)
    offset += poopt.wcswidth(level)
    color = msg.color
    PreMessageHelpers.write_nickname(win, level, color, False)
    win.addstr('> ')
    offset += 2
    return offset


@write_pre.register(Message)
def write_pre_message(msg: Message, win: Win, with_timestamps: bool, nick_size: int) -> int:
    """Write the part before the body:
        - timestamp (short or long)
        - ack/nack
        - nick (with a "* " for /me)
        - LMC number if present
    """
    color: TextStyle | None
    offset = 0
    if with_timestamps:
        offset += PreMessageHelpers.write_time(win, msg.history, msg.time)

    if not msg.nickname:  # not a message, nothing to do afterwards
        return offset

    nick = truncate_nick(msg.nickname, nick_size)
    offset += poopt.wcswidth(nick)
    if msg.nick_color:
        color = msg.nick_color
    elif msg.user:
        color = msg.user.color
    else:
        color = None
    if msg.ack:
        if msg.ack > 0:
            offset += PreMessageHelpers.write_ack(win)
        else:
            offset += PreMessageHelpers.write_nack(win)
    theme = get_theme()
    if msg.me:
        with win.colored_text(color=theme.COLOR_ME_MESSAGE):
            win.addstr(theme.CHAR_BEFORE_NICK_ME)
        PreMessageHelpers.write_nickname(win, nick, color, msg.highlight)
        offset += PreMessageHelpers.write_revisions(win, msg)
        win.addstr(theme.CHAR_AFTER_NICK_ME)
        offset += len(theme.CHAR_BEFORE_NICK_ME) + len(theme.CHAR_AFTER_NICK_ME)
    else:
        PreMessageHelpers.write_nickname(win, nick, color, msg.highlight)
        offset += PreMessageHelpers.write_revisions(win, msg)
        win.addstr(theme.CHAR_AFTER_NICK)
        offset += len(theme.CHAR_AFTER_NICK)
    return offset


@write_pre.register(XMLLog)
def write_pre_xmllog(msg: XMLLog, win: Win, with_timestamps: bool, nick_size: int) -> int:
    """Write the part before the stanza (timestamp + IN/OUT)"""
    offset = 0
    if with_timestamps:
        offset += 1 + PreMessageHelpers.write_time(win, False, msg.time)
    theme = get_theme()
    if msg.incoming:
        char = theme.CHAR_XML_IN
        color = theme.COLOR_XML_IN
    else:
        char = theme.CHAR_XML_OUT
        color = theme.COLOR_XML_OUT
    nick = truncate_nick(char, nick_size)
    offset += poopt.wcswidth(nick)
    PreMessageHelpers.write_nickname(win, char, color)
    win.addstr(' ')
    return offset


class PreMessageHelpers:
    """Collection of helpers to write the elements before a Message’s body."""

    @staticmethod
    def write_revisions(buffer: Win, msg: Message) -> int:
        """Write the number of message revisions, if any."""
        if msg.revisions:
            color = get_theme().COLOR_REVISIONS_MESSAGE
            with buffer.colored_text(color=color):
                buffer.addstr('%d' % msg.revisions)
            return ceil(log10(msg.revisions + 1))
        return 0

    @staticmethod
    def write_ack(buffer: Win) -> int:
        """Write the green checkmark for delivered messages."""
        theme = get_theme()
        color = theme.COLOR_CHAR_ACK
        with buffer.colored_text(color=color):
            buffer.addstr(theme.CHAR_ACK_RECEIVED)
        buffer.addstr(' ')
        return poopt.wcswidth(theme.CHAR_ACK_RECEIVED) + 1

    @staticmethod
    def write_nack(buffer: Win) -> int:
        """Write the red checkmark for errored messages."""
        theme = get_theme()
        color = theme.COLOR_CHAR_NACK
        with buffer.colored_text(color=color):
            buffer.addstr(theme.CHAR_NACK)
        buffer.addstr(' ')
        return poopt.wcswidth(theme.CHAR_NACK) + 1

    @staticmethod
    def write_nickname(buffer: Win, nickname: str, color: TextStyle | None, highlight: bool = False) -> None:
        """
        Write the nickname, using the user's color
        and return the number of written characters
        """
        if not nickname:
            return
        attr = None
        if highlight:
            theme_hl = get_theme().COLOR_HIGHLIGHT_NICK
            if theme_hl == "reverse":
                attr = curses.A_REVERSE
            elif isinstance(theme_hl, str):
                pass
            else:
                color = theme_hl
        with buffer.colored_text(color=color, attr=attr):
            buffer.addstr(nickname)

    @staticmethod
    def write_time(buffer: Win, history: bool, time: datetime) -> int:
        """
        Write the date on the yth line of the window
        """
        if time:
            theme = get_theme()
            if history and time.date() != date.today():
                format = theme.LONG_TIME_FORMAT
            else:
                format = theme.SHORT_TIME_FORMAT
            time_str = time.strftime(format)
            color = theme.COLOR_TIME_STRING
            with buffer.colored_text(color=color):
                buffer.addstr(time_str)
            buffer.addstr(' ')
            return poopt.wcswidth(time_str) + 1
        return 0



"""
Completions for the global commands
"""
import logging
import os
from functools import reduce
from pathlib import Path
from typing import List, TYPE_CHECKING

from slixmpp import JID, InvalidJID, Iq

from poezio import common
from poezio import tabs
from poezio import xdg
from poezio.config import config
from poezio.roster import roster

from poezio.core.structs import POSSIBLE_SHOW, Completion

if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.windows import Input

log = logging.getLogger(__name__)


class CompletionCore:
    def __init__(self, core: 'Core') -> None:
        self.core = core

    def help(self, the_input: 'Input') -> Completion:
        """Completion for /help."""
        commands = sorted(self.core.commands.keys()) + sorted(
            self.core.tabs.current_tab.commands.keys())
        return Completion(the_input.new_completion, commands, 1, quotify=False)

    def status(self, the_input: 'Input') -> Completion | None:
        """
        Completion of /status
        """
        if the_input.get_argument_position() == 1:
            return Completion(
                the_input.new_completion, [status for status in POSSIBLE_SHOW],
                1,
                ' ',
                quotify=False)
        return None

    def roster_barejids(self, the_input: 'Input') -> Completion:
        """Complete roster bare jids"""
        jids = sorted(
            str(contact.bare_jid) for contact in roster.contacts.values()
            if contact.pending_in
        )
        return Completion(the_input.new_completion, jids, 1, '', quotify=False)

    def remove(self, the_input: 'Input') -> Completion:
        """Completion for /remove"""
        jids = [str(jid) for jid in roster.jids()]
        return Completion(the_input.auto_completion, jids, '', quotify=False)

    def presence(self, the_input: 'Input') -> Completion | None:
        """
        Completion of /presence
        """
        arg = the_input.get_argument_position()
        if arg == 1:
            to_suggest = []
            for bookmark in self.core.bookmarks:
                tab = self.core.tabs.by_name_and_class(bookmark.jid,
                                                       tabs.MucTab)
                if tab is not None and tab.joined:
                    to_suggest.append(str(bookmark.jid))
            return Completion(
                the_input.auto_completion,
                [str(j) for j in roster.jids()] + to_suggest,
                '',
                quotify=True)
        elif arg == 2:
            return Completion(
                the_input.auto_completion,
                [status for status in POSSIBLE_SHOW],
                '',
                quotify=True)
        return None

    def theme(self, the_input: 'Input') -> Completion | None:
        """ Completion for /theme"""
        themes_dir_str = config.getstr('themes_dir')
        themes_dir = Path(themes_dir_str).expanduser(
        ) if themes_dir_str else xdg.DATA_HOME / 'themes'
        try:
            theme_files = [
                name.stem for name in themes_dir.iterdir()
                if name.suffix == '.py' and name.name != '__init__.py'
            ]
        except OSError:
            log.exception('Completion for /theme failed')
            return None
        if 'default' not in theme_files:
            theme_files.append('default')
        return Completion(
            the_input.new_completion, theme_files, 1, '', quotify=False)

    def win(self, the_input: 'Input') -> Completion:
        """Completion for /win"""
        comp_list = []
        for tab in self.core.tabs:
            comp_list.extend(tab.matching_names())
        comp = [i[1] for i in comp_list]
        return Completion(the_input.new_completion, comp, 1, '', quotify=False)

    def join(self, the_input: 'Input') -> Completion | None:
        """
        Completion for /join

        Try to complete the MUC JID:
            if only a resource is provided, complete with the default nick
            if only a server is provided, complete with the rooms from the
                disco#items of that server
            if only a nodepart is provided, complete with the servers of the
                current joined rooms
        """
        n = the_input.get_argument_position(quoted=True)
        args = common.shell_split(the_input.text)
        if n != 1:
            # we are not on the 1st argument of the command line
            return None
        if len(args) == 1:
            args.append('')
        try:
            jid = JID(args[1])
        except InvalidJID:
            jid = JID('')
            if args[1].endswith('@'):
                jid.user = args[1][:-1]

        relevant_rooms = []
        relevant_rooms.extend(sorted(self.core.pending_invites.keys()))
        bookmarks = [(str(elem.jid)
                      if not elem.nick else '%s/%s' % (elem.jid, elem.nick))
                     for elem in self.core.bookmarks]
        to_suggest = []
        for bookmark in bookmarks:
            tab = self.core.tabs.by_name_and_class(bookmark, tabs.MucTab)
            if not tab or (tab and not tab.joined):
                to_suggest.append(bookmark)
        relevant_rooms.extend(sorted(to_suggest))

        if the_input.last_completion:
            return Completion(the_input.new_completion, [], 1, quotify=True)

        if jid.user:
            # we are writing the server: complete the server
            serv_list = []
            for tab in self.core.get_tabs(tabs.MucTab):
                if tab.joined:
                    serv_list.append(
                        '%s@%s' % (jid.user, tab.general_jid.server)
                    )
            serv_list.extend(relevant_rooms)
            return Completion(
                the_input.new_completion, serv_list, 1, quotify=True)
        elif args[1].startswith('/'):
            # we completing only a resource
            return Completion(
                the_input.new_completion, ['/%s' % self.core.own_nick],
                1,
                quotify=True)
        else:
            return Completion(
                the_input.new_completion, relevant_rooms, 1, quotify=True)

    def version(self, the_input: 'Input') -> Completion:
        """Completion for /version"""
        comp: list[str] = reduce(
            lambda x, y: x + [str(i.jid) for i in y],
            (
                roster.contacts[jid].resources for jid in roster.jids()
                if len(roster.contacts[jid])
            ),
            []
        )
        return Completion(
            the_input.new_completion, sorted(comp), 1, quotify=False)

    def list(self, the_input: 'Input') -> Completion | None:
        """Completion for /list"""
        muc_serv_list = []
        for tab in self.core.get_tabs(
                tabs.MucTab):  # TODO, also from an history
            if tab.jid.server not in muc_serv_list:
                muc_serv_list.append(tab.jid.server)
        if muc_serv_list:
            return Completion(
                the_input.new_completion, muc_serv_list, 1, quotify=False)
        return None

    def move_tab(self, the_input: 'Input') -> Completion | None:
        """Completion for /move_tab"""
        n = the_input.get_argument_position(quoted=True)
        if n == 1:
            nodes = [tab.name for tab in self.core.tabs if tab]
            nodes.remove('Roster')
            return Completion(
                the_input.new_completion, nodes, 1, ' ', quotify=True)
        return None

    def runkey(self, the_input: 'Input') -> Completion:
        """
        Completion for /runkey
        """
        list_: list[str] = []
        list_.extend(self.core.key_func.keys())
        list_.extend(self.core.tabs.current_tab.key_func.keys())
        return Completion(the_input.new_completion, list_, 1, quotify=False)

    def bookmark(self, the_input: 'Input') -> Completion | None:
        """Completion for /bookmark"""
        args = common.shell_split(the_input.text)
        n = the_input.get_argument_position(quoted=True)

        if n == 2:
            return Completion(
                the_input.new_completion, ['true', 'false'], 2, quotify=True)
        if n >= 3:
            return None

        if len(args) == 1:
            args.append('')
        try:
            jid = JID(args[1])
        except InvalidJID:
            muc_list = [tab.name for tab in self.core.get_tabs(tabs.MucTab)]
            muc_list.sort()
            muc_list.append('*')
            return Completion(the_input.new_completion, muc_list, 1, quotify=True)
        tab = self.core.tabs.by_name_and_class(jid.bare, tabs.MucTab)
        nicks = [tab.own_nick] if tab else []
        default = os.environ.get('USER') if os.environ.get(
            'USER') else 'poezio'
        nick = config.getstr('default_nick') or ''
        if not nick:
            if default and default not in nicks:
                nicks.append(default)
        else:
            if nick not in nicks:
                nicks.append(nick)
        jids_list = ['%s/%s' % (jid.bare, nick) for nick in nicks]
        return Completion(
            the_input.new_completion, jids_list, 1, quotify=True)

    def remove_bookmark(self, the_input: 'Input') -> Completion:
        """Completion for /remove_bookmark"""
        return Completion(
            the_input.new_completion, [str(bm.jid) for bm in self.core.bookmarks],
            1,
            quotify=False)

    def decline(self, the_input: 'Input') -> Completion | None:
        """Completion for /decline"""
        n = the_input.get_argument_position(quoted=True)
        if n == 1:
            return Completion(
                the_input.auto_completion,
                sorted(self.core.pending_invites.keys()),
                '',
                quotify=True)
        return None

    def bind(self, the_input: 'Input') -> Completion | None:
        n = the_input.get_argument_position()
        if n == 1:
            args = [
                key for key in self.core.key_func if not key.startswith('_')
            ]
        elif n == 2:
            args = [key for key in self.core.key_func]
        else:
            return None

        return Completion(the_input.new_completion, args, n, '', quotify=False)

    def message(self, the_input: 'Input') -> Completion | None:
        """Completion for /message"""
        n = the_input.get_argument_position(quoted=True)
        if n >= 2:
            return None
        online: list[str] = []
        offline: list[str] = []
        for jid in sorted((str(jid) for jid in roster.jids())):
            if len(roster.contacts[jid]) > 0:
                online.append(jid)
            else:
                offline.append(jid)
        return Completion(
            the_input.new_completion, online + offline, 1, '', quotify=True)

    def invite(self, the_input: 'Input') -> Completion | None:
        """Completion for /invite"""
        n = the_input.get_argument_position(quoted=True)
        if n == 1:
            comp: list[str] = reduce(
                lambda x, y: x + [str(i.jid) for i in y],
                (
                    roster.contacts[jid].resources for jid in roster.jids()
                    if len(roster.contacts[jid])
                ),
                [],
            )
            comp = sorted(comp)
            bares = sorted(
                str(roster.contacts[contact].bare_jid)
                for contact in roster.jids()
                if len(roster.contacts[contact])
            )
            off = sorted(str(jid) for jid in roster.jids() if jid not in bares)
            comp = comp + bares + off
            return Completion(the_input.new_completion, comp, n, quotify=True)
        elif n == 2:
            rooms = []
            for tab in self.core.get_tabs(tabs.MucTab):
                if tab.joined:
                    rooms.append(tab.jid.bare)
            rooms.sort()
            return Completion(
                the_input.new_completion, rooms, n, '', quotify=True)
        return None

    def impromptu(self, the_input: 'Input') -> Completion:
        """Completion for /impromptu"""
        n = the_input.get_argument_position(quoted=True)
        onlines = []
        offlines = []
        for barejid in roster.jids():
            if len(roster.contacts[barejid]):
                onlines.append(str(barejid))
            else:
                offlines.append(str(barejid))
        comp = sorted(onlines) + sorted(offlines)
        return Completion(the_input.new_completion, comp, n, quotify=True)

    def last_activity(self, the_input: 'Input') -> Completion | None:
        """
        Completion for /last_activity <jid>
        """
        n = the_input.get_argument_position(quoted=False)
        if n >= 2:
            return None
        comp: list[str] = reduce(
            lambda x, y: x + [str(i.jid) for i in y],
            (
                roster.contacts[jid].resources for jid in roster.jids()
                if len(roster.contacts[jid])
            ),
            [],
        )
        return Completion(
            the_input.new_completion, sorted(comp), 1, '', quotify=False)

    def server_cycle(self, the_input: 'Input') -> Completion:
        """Completion for /server_cycle"""
        serv_list = set()
        for tab in self.core.get_tabs(tabs.MucTab):
            serv_list.add(tab.jid.server)
        return Completion(the_input.new_completion, sorted(serv_list), 1, ' ')

    def set(self, the_input: 'Input') -> Completion | None:
        """Completion for /set"""
        args = common.shell_split(the_input.text)
        n = the_input.get_argument_position(quoted=True)
        if n >= len(args):
            args.append('')
        if n == 1:
            if '|' in args[1]:
                plugin_name, section = args[1].split('|')[:2]
                if plugin_name not in self.core.plugin_manager.plugins:
                    return Completion(
                        the_input.new_completion, [], n, quotify=True)
                plugin = self.core.plugin_manager.plugins[plugin_name]
                if plugin is None:
                    return None
                end_list = [
                    '%s|%s' % (plugin_name, section)
                    for section in plugin.config.sections()
                ]
            else:
                end_set = set(config.options('Poezio'))
                end_set.update(config.default.get('Poezio', {}))
                end_list = list(end_set)
                end_list.sort()
        elif n == 2:
            if '|' in args[1]:
                plugin_name, section = args[1].split('|')[:2]
                if plugin_name not in self.core.plugin_manager.plugins:
                    return Completion(
                        the_input.new_completion, [''], n, quotify=True)
                plugin = self.core.plugin_manager.plugins[plugin_name]
                if plugin is None:
                    return None
                end_set = set(plugin.config.options(section or plugin_name))
                if plugin.config.default:
                    end_set.update(
                        plugin.config.default.get(section or plugin_name, {}))
                end_list = list(end_set)
                end_list.sort()
            elif not config.has_option('Poezio', args[1]):
                if config.has_section(args[1]):
                    end_list = config.options(args[1])
                    end_list.append('')
                else:
                    end_list = []
            else:
                end_list = [str(config.get(args[1], '')), '']
        elif n == 3:
            if '|' in args[1]:
                plugin_name, section = args[1].split('|')[:2]
                if plugin_name not in self.core.plugin_manager.plugins:
                    return Completion(
                        the_input.new_completion, [''], n, quotify=True)
                plugin = self.core.plugin_manager.plugins[plugin_name]
                if plugin is None:
                    return None
                end_list = [
                    str(
                        plugin.config.get(args[2], '', section
                                          or plugin_name)), ''
                ]
            else:
                if not config.has_section(args[1]):
                    end_list = ['']
                else:
                    end_list = [str(config.get(args[2], '', args[1])), '']
        else:
            return None
        return Completion(the_input.new_completion, end_list, n, quotify=True)

    def set_default(self, the_input: 'Input') -> Completion | None:
        """ Completion for /set_default
        """
        args = common.shell_split(the_input.text)
        n = the_input.get_argument_position(quoted=True)
        if n >= len(args):
            args.append('')
        if n == 1 or (n == 2 and config.has_section(args[1])):
            return self.set(the_input)
        return None

    def toggle(self, the_input: 'Input') -> Completion:
        "Completion for /toggle"
        return Completion(
            the_input.new_completion,
            config.options('Poezio'),
            1,
            quotify=False)

    def bookmark_local(self, the_input: 'Input') -> Completion | None:
        """Completion for /bookmark_local"""
        n = the_input.get_argument_position(quoted=True)
        args = common.shell_split(the_input.text)

        if n >= 2:
            return None
        if len(args) == 1:
            args.append('')
        try:
            jid = JID(args[1])
        except InvalidJID:
            muc_list = [tab.name for tab in self.core.get_tabs(tabs.MucTab)]
            muc_list.append('*')
            return Completion(the_input.new_completion, muc_list, 1, quotify=True)
        tab = self.core.tabs.by_name_and_class(jid.bare, tabs.MucTab)
        nicks = [tab.own_nick] if tab else []
        default = os.environ.get('USER') if os.environ.get(
            'USER') else 'poezio'
        nick = config.getstr('default_nick')
        if not nick:
            if default and default not in nicks:
                nicks.append(default)
        else:
            if nick not in nicks:
                nicks.append(nick)
        jids_list = ['%s/%s' % (jid.bare, nick) for nick in nicks]
        return Completion(
            the_input.new_completion, jids_list, 1, quotify=True)

    def block(self, the_input: 'Input') -> Completion | None:
        """
        Completion for /block
        """
        if the_input.get_argument_position() == 1:

            current_tab = self.core.tabs.current_tab
            chattabs = (
                tabs.ConversationTab,
                tabs.StaticConversationTab,
                tabs.DynamicConversationTab,
            )
            tabjid: List[str] = []
            if isinstance(current_tab, chattabs):
                tabjid = [current_tab.jid.bare]

            jids = [str(i) for i in roster.jids()]
            jids += tabjid
            return Completion(
                the_input.new_completion, jids, 1, '', quotify=False)
        return None

    def unblock(self, the_input: 'Input') -> Completion | None:
        """
        Completion for /unblock
        """

        def on_result(iq: Iq) -> Completion | None:
            if iq['type'] == 'error':
                return None
            comp = sorted(str(item['jid']) for item in iq['blocklist']['items'])
            return Completion(the_input.new_completion, comp, 1, quotify=False)

        if the_input.get_argument_position():
            self.core.xmpp.plugin['xep_0191'].get_blocked(callback=on_result)
        return None

    report = block



"""
Display the time between two messages.

Helps you identify the times of a conversation.  For example
if you disable the timestamps, and remove the join/quit notifications in a
chatroom, you can’t really distinguish when a conversation stopped and when
a new one started, because you don’t have a visual separation between the two.

This plugin displays a message in the conversation indicating the time that
passed between two messages, if the time is bigger than X minutes
(configurable, of course. Default is 15 minutes). This way you know how many time elapsed between
them, letting you understand more easily what is going on without any visual
clutter.

Configuration
-------------

You can configure the minimum delay between two messages, to display the time marker, in seconds. The default is 10 minutes (aka 600 seconds).

.. code-block:: ini

    [time_marker]
    delay = 600

Usage
-----

Messages like “2 hours, 25 minutes passed…” are automatically displayed into the conversation. You don’t need to (and can’t) do anything.

"""

from poezio.plugin import BasePlugin
from poezio import tabs
from typing import Any
from datetime import datetime, timedelta
from poezio.ui.types import InfoMessage


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler("muc_msg", self.on_muc_msg)
        # Dict of MucTab.jid.bare: last_message date, so we don’t have to
        # retrieve the messages of the given muc to look for the last
        # message’s date each time.  Also, now that I think about it, the
        # date of the message is not event kept in the Message object, so…
        self.last_messages: dict[str, datetime] = {}

    def on_muc_msg(self, message: Any, tab: tabs.MucTab) -> None:
        def format_timedelta(delta: timedelta) -> str:
            """
            Return a string of the form D days, H hours, M minutes, S
            seconds.  If the number of total minutes is bigger than 10, we
            usually don’t care anymore about the number of seconds, so we
            don’t display it.  Same thing if the number of days is bigger
            than one, we don’t display the minutes either.
            """
            days = delta.days
            hours = delta.seconds // 3600
            minutes = delta.seconds // 60 % 60
            seconds = delta.seconds % 60
            res = ''
            if days > 0:
                res = "%s days, " % days
            if hours > 0:
                res += "%s hours, " % hours
            if days == 0 and minutes != 0:
                res += "%s minutes, " % minutes
            if delta.total_seconds() < 600:
                res += "%s seconds, " % seconds
            return res[:-2]

        last_message_date = self.last_messages.get(tab.jid.bare)
        self.last_messages[tab.jid.bare] = datetime.now()
        if last_message_date:
            delta = datetime.now() - last_message_date
            if delta >= timedelta(0, self.config.get('delay', 900)):
                tab.add_message(
                    InfoMessage("%s passed…" % (format_timedelta(delta), ))
                )



"""
Do a disco#info query on a JID, display the XEP-0157 Contact Addresses

Usage
-----

.. glossary::

    /contact
        **Usage:** ``/contact <JID>``

        This command queries a JID for its Contact Addresses.
"""

from poezio.plugin import BasePlugin
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp import InvalidJID, Iq, JID

CONTACT_TYPES = ['abuse', 'admin', 'feedback', 'sales', 'security', 'support']


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'contact',
            self.command_disco,
            usage='<JID>',
            short='Get the Contact Addresses of a JID',
            help='Get the Contact Addresses of a JID',
        )

    def on_disco(self, iq: Iq) -> None:
        contacts = []
        # iterate all data forms, in case there are multiple
        for form in iq['disco_info']:
            values = form.get_values()
            if values['FORM_TYPE'][0] == 'http://jabber.org/network/serverinfo':
                for var in values:
                    if not var.endswith('-addresses'):
                        continue
                    title = var[:-10]  # strip '-addresses'
                    sep = '\n  ' + len(title) * ' '
                    field_value = values[var]
                    if field_value:
                        value = sep.join(field_value) if isinstance(field_value, list) else field_value
                        contacts.append(f'{title}: {value}')
        if contacts:
            self.api.information('\n'.join(contacts), 'Contact Info')
        else:
            self.api.information(f'No Contact Addresses for {iq["from"]}', 'Error')

    async def command_disco(self, jid: str) -> None:
        try:
            iq = await self.core.xmpp.plugin['xep_0030'].get_info(jid=JID(jid), cached=False)
            self.on_disco(iq)
        except InvalidJID as exn:
            self.api.information(f'Invalid JID “{jid}”: {exn}', 'Error')
        except (IqError, IqTimeout,) as exn:
            ifrom = exn.iq['from']
            condition = exn.iq['error']['condition']
            text = exn.iq['error']['text']
            message = f'Error getting Contact Addresses from {ifrom}: {condition}: {text}'
            self.api.information(message, 'Error')



"""
Double the first word of any message you send in a :ref:`muctab`, making you appear retarded.
"""
from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.double)

    def double(self, msg: Message, tab: Any) -> None:
        split = msg['body'].split()
        if split:
            msg['body'] = split[0] + ' ' + msg['body']



#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Defines the EventHandler class.
The list of available events is here:
http://poezio.eu/doc/en/plugins.html#_poezio_events
"""
import logging

from collections import OrderedDict
from inspect import iscoroutinefunction
from typing import Callable, List, Any

log = logging.getLogger(__name__)


class EventHandler:
    """
    A class keeping a list of possible events that are triggered
    by poezio. You (a plugin for example) can add an event handler
    associated with an event name, and whenever that event is triggered,
    the callback is called.
    """

    def __init__(self) -> None:
        events = [
            'highlight',
            'muc_say',
            'muc_say_after',
            'conversation_say',
            'conversation_say_after',
            'private_say',
            'private_say_after',
            'conversation_msg',
            'private_msg',
            'muc_msg',
            'conversation_chatstate',
            'muc_chatstate',
            'private_chatstate',
            'normal_presence',
            'muc_presence',
            'muc_join',
            'joining_muc',
            'changing_nick',
            'muc_kick',
            'muc_nickchange',
            'muc_ban',
            'send_normal_presence',
            'ignored_private',
            'tab_change',
        ]
        self.events: dict[str, OrderedDict[int, List[Callable[..., Any]]]] = {}
        for event in events:
            self.events[event] = OrderedDict()

    def add_event_handler(self, name: str, callback: Callable[..., Any],
                          priority: int = 50) -> bool:
        """
        Add a callback to a given event.
        Note that if that event name doesn’t exist, it just returns False.
        If it was successfully added, it returns True
        priority is a integer between 0 and 100. 0 is the highest priority and
        will be called first. 100 is the lowest.
        """

        if name not in self.events:
            return False

        callbacks = self.events[name]

        # Clamp priority
        priority = max(0, min(priority, 100))

        entry = callbacks.setdefault(priority, [])
        entry.append(callback)

        return True

    async def trigger_async(self, name: str, *args: Any, **kwargs: Any) -> None:
        """
        Call all the callbacks associated to the given event name.
        """
        callbacks = self.events.get(name, None)
        if callbacks is None:
            return
        for priority in callbacks.values():
            for callback in priority:
                if iscoroutinefunction(callback):
                    await callback(*args, **kwargs)
                else:
                    callback(*args, **kwargs)

    def trigger(self, name: str, *args: Any, **kwargs: Any) -> None:
        """
        Call all the callbacks associated to the given event name.
        """
        callbacks = self.events.get(name, None)
        if callbacks is None:
            return
        for priority in callbacks.values():
            for callback in priority:
                if not iscoroutinefunction(callback):
                    callback(*args, **kwargs)
                else:
                    log.error(
                        f'async event handler {callback} '
                        'called in sync trigger!'
                    )

    def del_event_handler(self, name: str | None, callback: Callable[..., Any]) -> None:
        """
        Remove the callback from the list of callbacks of the given event
        """
        if not name:
            for callbacks in self.events.values():
                for priority in callbacks.values():
                    for entry in priority[:]:
                        if entry == callback:
                            priority.remove(callback)
        else:
            callbacks = self.events[name]
            for priority in callbacks.values():
                for entry in priority[:]:
                    if entry == callback:
                        priority.remove(callback)



"""
Tests for the Tabs list module
"""

from poezio.core.tabs import Tabs
from poezio.tabs import GapTab, Tab
from poezio.events import EventHandler

h = EventHandler()

class DummyTab(Tab):
    count = 0

    def __init__(self):
        self._name = 'dummy%s' % self.count
        DummyTab.count += 1

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @staticmethod
    def reset():
        DummyTab.count = 0

def test_append():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    tabs.append(dummy)
    assert tabs[0] is dummy
    assert tabs[0].nb == 0
    assert tabs['dummy0'] is dummy
    assert tabs.by_class(DummyTab) == [dummy]

    dummy2 = DummyTab()
    tabs.append(dummy2)
    assert tabs[1] is dummy2
    assert tabs[1].nb == 1
    assert tabs['dummy1'] is dummy2
    assert tabs.by_class(DummyTab) == [dummy, dummy2]

def test_delete():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)

    tabs.delete(dummy)
    assert tabs[0] is dummy2
    assert tabs.by_class(DummyTab) == [dummy2]
    assert tabs['dummy0'] is None
    assert dummy2.nb == 0


def test_delete_restore_previous():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    dummy4 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)
    tabs.append(dummy4)

    tabs.set_current_tab(dummy4)
    tabs.set_current_tab(dummy2)
    tabs.delete(dummy2)
    assert tabs.current_tab is dummy4
    assert tabs.current_index == 2
    assert dummy.nb == 0
    assert dummy3.nb == 1
    assert dummy4.nb == 2

def test_delete_other_tab():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    dummy4 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)
    tabs.append(dummy4)

    tabs.set_current_tab(dummy4)
    tabs.delete(dummy2)
    assert tabs.current_tab is dummy4
    assert tabs.current_index == 2
    assert dummy.nb == 0
    assert dummy3.nb == 1
    assert dummy4.nb == 2

def test_insert_and_gaps():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)

    tabs.insert_tab(2, 1)
    assert tabs[1] is dummy3
    assert tabs[2] is dummy2

    tabs.insert_tab(1, 100, gaps=True)
    assert isinstance(tabs[1], GapTab)
    assert tabs[3] is dummy3
    assert tabs[3].nb == 3

    tabs.update_gaps(enable_gaps=False)
    assert tabs[1] is dummy2
    assert tabs[1].nb == 1

def test_replace_tabs():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)

    tabs.replace_tabs([dummy3, dummy2, dummy])
    assert tabs[0] is dummy3
    assert tabs[2] is dummy
    assert tabs[2].nb == 2

def test_prev_next():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)

    for idx in range(6):
        assert tabs.current_index == idx % 3
        tabs.next()

    for idx in range(6):
        assert (3 - idx) % 3 == tabs.current_index
        tabs.prev()

    tabs.insert_tab(1, 999, gaps=True)

    assert tabs.current_index == 0
    tabs.next()
    assert tabs.current_index == 2
    tabs.prev()
    assert tabs.current_index == 0

def test_set_current():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)

    assert tabs.current_tab is dummy
    tabs.set_current_index(2)
    assert tabs.current_tab is dummy3
    tabs.set_current_tab(dummy2)
    assert tabs.current_tab is dummy2

def test_slice():
    DummyTab.reset()
    tabs = Tabs(h, None)
    dummy = DummyTab()
    dummy2 = DummyTab()
    dummy3 = DummyTab()
    tabs.append(dummy)
    tabs.append(dummy2)
    tabs.append(dummy3)

    assert tabs[1:2][0] is dummy2

def test_find_by_unique_prefix():
    DummyTab.reset()
    tabs = Tabs(h, None)
    t1 = DummyTab()
    t2 = DummyTab()
    t3 = DummyTab()
    tabs.append(t1)
    tabs.append(t2)
    tabs.append(t3)

    t1.name = "foo"
    t2.name = "bar"
    t3.name = "fnord"

    assert tabs.find_by_unique_prefix("f") == (True, None)
    assert tabs.find_by_unique_prefix("b") == (True, t2)
    assert tabs.find_by_unique_prefix("fo") == (True, t1)
    assert tabs.find_by_unique_prefix("fn") == (True, t3)
    assert tabs.find_by_unique_prefix("fx") == (False, None)
    assert tabs.find_by_unique_prefix("x") == (False, None)
    assert tabs.find_by_unique_prefix("") == (True, None)



"""
Windows used with the roster (window displaying the contacts, and the
one showing detailed info on the current selection)
"""
import logging
from datetime import datetime
from typing import Optional, List, Union

from poezio.windows.base_wins import Win

from poezio.config import config
from poezio.contact import Contact, Resource
from poezio.roster import Roster, RosterGroup
from poezio.theming import get_theme, to_curses_attr

log = logging.getLogger(__name__)

Row = Union[RosterGroup, Contact, Resource]


class RosterWin(Win):
    __slots__ = ('pos', 'start_pos', 'selected_row', 'roster_cache')

    def __init__(self) -> None:
        Win.__init__(self)
        self.pos = 0  # cursor position in the contact list
        self.start_pos = 1  # position of the start of the display
        self.selected_row: Optional[Row] = None
        self.roster_cache: List[Row] = []

    @property
    def roster_len(self) -> int:
        return len(self.roster_cache)

    def move_cursor_down(self, number: int = 1) -> bool:
        """
        Return True if we scrolled, False otherwise
        """
        pos = self.pos
        if self.pos < self.roster_len - number:
            self.pos += number
        else:
            self.pos = self.roster_len - 1
        if self.pos >= self.start_pos - 1 + self.height - 1:
            if number == 1:
                self.scroll_down(8)
            else:
                self.scroll_down(self.pos - self.start_pos - self.height // 2)
        self.update_pos()
        return pos != self.pos

    def move_cursor_up(self, number: int = 1) -> bool:
        """
        Return True if we scrolled, False otherwise
        """
        pos = self.pos
        if self.pos - number >= 0:
            self.pos -= number
        else:
            self.pos = 0
        if self.pos <= self.start_pos:
            if number == 1:
                self.scroll_up(8)
            else:
                self.scroll_up(self.start_pos - self.pos + self.height // 2)
        self.update_pos()
        return pos != self.pos

    def update_pos(self) -> None:
        if len(self.roster_cache) > self.pos and self.pos >= 0:
            self.selected_row = self.roster_cache[self.pos]
        elif self.roster_cache:
            self.selected_row = self.roster_cache[-1]

    def scroll_down(self, number: int = 8) -> bool:
        pos = self.start_pos
        if self.start_pos + number <= self.roster_len - 1:
            self.start_pos += number
        else:
            self.start_pos = self.roster_len - 1
        return self.start_pos != pos

    def scroll_up(self, number: int = 8) -> bool:
        pos = self.start_pos
        if self.start_pos - number > 0:
            self.start_pos -= number
        else:
            self.start_pos = 1
        return self.start_pos != pos

    def build_roster_cache(self, roster: Roster) -> None:
        """
        Regenerates the roster cache if needed
        """
        if not roster.needs_rebuild:
            return
        log.debug('The roster has changed, rebuilding the cache…')
        # This is a search
        if roster.contact_filter is not roster.DEFAULT_FILTER:
            self.roster_cache = []
            sort = config.getstr('roster_sort') or 'jid:show'
            for contact in roster.get_contacts_sorted_filtered(sort):
                self.roster_cache.append(contact)
        else:
            show_offline = config.getbool('roster_show_offline')
            sort = config.getstr('roster_sort') or 'jid:show'
            group_sort = config.getstr('roster_group_sort') or 'name'
            self.roster_cache = []
            # build the cache
            for group in roster.get_groups(group_sort):
                contacts_filtered = group.get_contacts()
                if (not show_offline and group.get_nb_connected_contacts() == 0)\
                        or not contacts_filtered:
                    continue  # Ignore empty groups
                self.roster_cache.append(group)
                if group.folded:
                    continue  # ignore folded groups
                for contact in group.get_contacts(sort=sort):
                    if not show_offline and len(contact) == 0:
                        continue  # ignore offline contacts
                    self.roster_cache.append(contact)
                    if not contact.folded(group.name):
                        for resource in contact.get_resources():
                            self.roster_cache.append(resource)
        roster.last_built = datetime.now()
        if self.selected_row in self.roster_cache:
            if self.pos < self.roster_len and self.roster_cache[self.
                                                                pos] != self.selected_row:
                self.pos = self.roster_cache.index(self.selected_row)

    def refresh(self, roster: Roster) -> None:
        """
        We display a number of lines from the roster cache
        (and rebuild it if needed)
        """
        log.debug('Refresh: %s', self.__class__.__name__)
        self.build_roster_cache(roster)
        # make sure we are within bounds
        self.move_cursor_up((
            self.roster_len + self.pos) if self.pos >= self.roster_len else 0)
        if not self.roster_cache:
            self.selected_row = None
        self._win.erase()
        self._win.move(0, 0)
        self.draw_roster_information(roster)
        y = 1
        group = "none"
        # scroll down if needed
        if self.start_pos + self.height <= self.pos + 2:
            self.scroll_down(self.pos - self.start_pos - self.height +
                             (self.height // 2))
        # draw the roster from the cache
        roster_view = self.roster_cache[self.start_pos - 1:self.start_pos +
                                        self.height]

        options = {
            'show_roster_sub': config.getbool('show_roster_subscriptions'),
            'show_s2s_errors': config.getbool('show_s2s_errors'),
            'show_roster_jids': config.getbool('show_roster_jids')
        }

        for item in roster_view:
            draw_selected = False
            if y - 2 + self.start_pos == self.pos:
                draw_selected = True
                self.selected_row = item

            if isinstance(item, RosterGroup):
                self.draw_group(y, item, draw_selected)
                group = item.name
            elif isinstance(item, Contact):
                self.draw_contact_line(y, item, draw_selected, group,
                                       **options)  # type: ignore
            elif isinstance(item, Resource):
                self.draw_resource_line(y, item, draw_selected)

            y += 1

        if self.start_pos > 1:
            self.draw_plus(1)
        if self.start_pos + self.height - 2 < self.roster_len:
            self.draw_plus(self.height - 1)
        self._refresh()

    def draw_plus(self, y: int) -> None:
        """
        Draw the indicator that shows that
        the list is longer than what is displayed
        """
        self.addstr(y, self.width - 5, '++++',
                    to_curses_attr(get_theme().COLOR_MORE_INDICATOR))

    def draw_roster_information(self, roster: Roster) -> None:
        """
        The header at the top
        """
        color = get_theme().COLOR_INFORMATION_BAR
        self.addstr(
            'Roster: %s/%s contacts' % (roster.get_nb_connected_contacts(),
                                        len(roster)),
            to_curses_attr(color))
        self.finish_line(color)

    def draw_group(self, y: int, group: RosterGroup, colored: bool) -> None:
        """
        Draw a groupname on a line
        """
        color = to_curses_attr(get_theme().COLOR_SELECTED_ROW)
        if colored:
            self._win.attron(color)
        if group.folded:
            self.addstr(y, 0, '[+] ')
        else:
            self.addstr(y, 0, '[-] ')
        contacts = " (%s/%s)" % (group.get_nb_connected_contacts(), len(group))
        self.addstr(
            y, 4,
            self.truncate_name(group.name,
                               len(contacts) + 4) + contacts)
        if colored:
            self._win.attroff(color)
        self.finish_line()

    def truncate_name(self, name: str, added: int) -> str:
        if len(name) + added <= self.width:
            return name
        return name[:self.width - added - 1] + '…'

    def draw_contact_line(self,
                          y: int,
                          contact: Contact,
                          colored: bool,
                          group: str,
                          show_roster_sub: str | None = None,
                          show_s2s_errors: bool = True,
                          show_roster_jids: bool = False) -> None:
        """
        Draw on a line all information about one contact.
        This is basically the highest priority resource's information
        Use 'color' to draw the jid/display_name to show what is
        the currently selected contact in the list
        """

        theme = get_theme()
        resource = contact.get_highest_priority_resource()
        if not resource:
            # There's no online resource
            presence = 'unavailable'
            nb = ''
        else:
            presence = resource.presence
            nb = ' (%s)' % len(contact)
        color = theme.color_show(presence)
        added = 2 + len(theme.CHAR_STATUS) + len(nb)

        self.addstr(y, 0, ' ')
        self.addstr(theme.CHAR_STATUS, to_curses_attr(color))

        self.addstr(' ')
        if resource:
            self.addstr('[+] ' if contact.folded(group) else '[-] ')
        else:
            self.addstr('    ')
        added += 4

        if contact.ask:
            added += len(theme.CHAR_ROSTER_ASKED)
        if show_s2s_errors and contact.error:
            added += len(theme.CHAR_ROSTER_ERROR)
        if show_roster_sub in ('all', 'incomplete', 'to', 'from', 'both',
                               'none'):
            added += len(
                theme.char_subscription(
                    contact.subscription, keep=show_roster_sub))

        if not show_roster_jids and contact.name:
            display_name = contact.name
        elif contact.name and contact.name != contact.bare_jid:
            display_name = '%s (%s)' % (contact.name, contact.bare_jid)
        else:
            display_name = str(contact.bare_jid)

        display_name = self.truncate_name(display_name, added) + nb

        if colored:
            self.addstr(display_name,
                        to_curses_attr(theme.COLOR_SELECTED_ROW))
        else:
            self.addstr(display_name)

        if show_roster_sub in ('all', 'incomplete', 'to', 'from', 'both',
                               'none'):
            self.addstr(
                theme.char_subscription(
                    contact.subscription, keep=show_roster_sub),
                to_curses_attr(theme.COLOR_ROSTER_SUBSCRIPTION))
        if contact.ask:
            self.addstr(theme.CHAR_ROSTER_ASKED,
                        to_curses_attr(theme.COLOR_IMPORTANT_TEXT))
        if show_s2s_errors and contact.error:
            self.addstr(theme.CHAR_ROSTER_ERROR,
                        to_curses_attr(theme.COLOR_ROSTER_ERROR))
        self.finish_line()

    def draw_resource_line(self, y: int, resource: Resource, colored: bool) -> None:
        """
        Draw a specific resource line
        """
        theme = get_theme()
        color = theme.color_show(resource.presence)
        self.addstr(y, 4, theme.CHAR_STATUS, to_curses_attr(color))
        if colored:
            self.addstr(y, 8, self.truncate_name(str(resource.jid), 6),
                        to_curses_attr(theme.COLOR_SELECTED_ROW))
        else:
            self.addstr(y, 8, self.truncate_name(str(resource.jid), 6))
        self.finish_line()

    def get_selected_row(self) -> Optional[Row]:
        if self.pos >= len(self.roster_cache):
            return self.selected_row
        if len(self.roster_cache) > 0:
            self.selected_row = self.roster_cache[self.pos]
            return self.roster_cache[self.pos]
        return None


class ContactInfoWin(Win):
    __slots__ = ()

    def draw_contact_info(self, contact: Contact) -> None:
        """
        draw the contact information
        """
        theme = get_theme()
        resource = contact.get_highest_priority_resource()
        if contact:
            jid = str(contact.bare_jid)
        elif resource:
            jid = str(resource.jid)
        else:
            jid = 'example@example.com'  # should never happen
        if resource:
            presence = resource.presence
        else:
            presence = 'unavailable'
        i = 0
        self.addstr(0, 0, '%s (%s)' % (
            jid,
            presence,
        ), to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self.finish_line(theme.COLOR_INFORMATION_BAR)
        i += 1
        self.addstr(i, 0, 'Subscription: %s' % (contact.subscription, ))
        self.finish_line()
        i += 1
        if contact.ask:
            if contact.ask == 'asked':
                self.addstr(i, 0, 'Ask: %s' % (contact.ask, ),
                            to_curses_attr(theme.COLOR_IMPORTANT_TEXT))
            else:
                self.addstr(i, 0, 'Ask: %s' % (contact.ask, ))
            self.finish_line()
            i += 1
        if resource:
            self.addstr(i, 0, 'Status: %s' % (resource.status))
            self.finish_line()
            i += 1

        if contact.error:
            self.addstr(i, 0, 'Error: %s' % contact.error,
                        to_curses_attr(theme.COLOR_ROSTER_ERROR))
            self.finish_line()
            i += 1

    def draw_group_info(self, group: RosterGroup) -> None:
        """
        draw the group information
        """
        theme = get_theme()
        self.addstr(0, 0, group.name,
                    to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self.finish_line(theme.COLOR_INFORMATION_BAR)

    def refresh(self, selected_row: Row) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        if isinstance(selected_row, RosterGroup):
            self.draw_group_info(selected_row)
        elif isinstance(selected_row, Contact):
            self.draw_contact_info(selected_row)
        # elif isinstance(selected_row, Resource):
        #     self.draw_contact_info(None, selected_row)
        self._refresh()



'''
This plugin lets the user select and send a sticker from a pack of stickers.

The protocol used here is based on XEP-0363 and XEP-0066, while a future
version may use XEP-0449 instead.

Command
-------

.. glossary::
    /sticker
        **Usage:** ``/sticker <pack>``

        Opens a picker tool, and send the sticker which has been selected.

Configuration options
---------------------

.. glossary::
    sticker_picker
        **Default:** ``poezio-sticker-picker``

        The command to invoke as a sticker picker.  A sample one is provided in
        tools/sticker-picker.

    stickers_dir
        **Default:** ``XDG_DATA_HOME/poezio/stickers``

        The directory under which the sticker packs can be found.
'''

import asyncio
from poezio import xdg, tabs
from poezio.plugin import BasePlugin
from poezio.config import config
from poezio.core.structs import Completion
from poezio.windows.inputs import Input
from pathlib import Path
from asyncio.subprocess import PIPE


class Plugin(BasePlugin):
    dependencies = {'upload'}

    def init(self) -> None:
        # The command to use as a picker helper.
        self.picker_command = config.getstr('sticker_picker') or 'poezio-sticker-picker'

        # Select and create the stickers directory.
        directory = config.getstr('stickers_dir')
        if directory:
            self.directory = Path(directory).expanduser()
        else:
            self.directory = xdg.DATA_HOME / 'stickers'
        self.directory.mkdir(parents=True, exist_ok=True)

        self.upload = self.refs['upload']
        self.api.add_command(
            'sticker',
            self.command_sticker,
            usage='<sticker pack>',
            short='Send a sticker',
            help='Send a sticker, with a helper GUI sticker picker',
            completion=self.completion_sticker,
        )

    def command_sticker(self, pack: str) -> None:
        '''
        Sends a sticker
        '''
        if not pack:
            self.api.information('Missing sticker pack argument.', 'Error')
            return

        async def run_command(tab: tabs.ChatTab, path: Path) -> None:
            try:
                process = await asyncio.create_subprocess_exec(
                    self.picker_command, path, stdout=PIPE, stderr=PIPE)
                sticker, stderr = await process.communicate()
            except FileNotFoundError as err:
                self.api.information('Failed to launch the sticker picker: %s' % err, 'Error')
                return
            else:
                if process.returncode != 0:
                    self.api.information('Sticker picker failed: %s' % stderr.decode(), 'Error')
                    return
            if sticker:
                filename = sticker.decode().rstrip()
                self.api.information('Sending sticker %s' % filename, 'Info')
                await self.upload.send_upload(path / filename, tab)
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        path = self.directory / pack
        self.core.register_task(run_command(tab, path))

    def completion_sticker(self, the_input: Input) -> Completion:
        '''
        Completion for /sticker
        '''
        txt = the_input.get_text()[9:]
        directories = [directory.name for directory in self.directory.glob(txt + '*')]
        return Completion(the_input.auto_completion, directories, quotify=False)



from typing import Tuple, Dict, List, Union
import curses
import hashlib

from . import hsluv

Palette = Dict[float, int]

# BT.601 (YCbCr) constants, see XEP-0392
K_R = 0.299
K_G = 0.587
K_B = 1 - K_R - K_G


class ColorNotFoundInPalette(ValueError):
    def __init__(self) -> None:
        super().__init__("Color not found in palette")


def ncurses_color_to_rgb(color: int) -> Tuple[float, float, float]:
    if color <= 15:
        r: Union[int, float]
        g: Union[int, float]
        b: Union[int, float]
        try:
            (r, g, b) = curses.color_content(color)
        except Exception:  # fallback in faulty terminals (e.g. xterm)
            (r, g, b) = curses.color_content(color % 8)
        r = r / 1000 * 5
        g = g / 1000 * 5
        b = b / 1000 * 5
    elif color <= 231:
        color = color - 16
        b = color % 6
        color = color // 6
        g = color % 6
        color = color // 6
        r = color % 6
    else:
        color -= 232
        r = g = b = color / 24 * 5
    return r / 5, g / 5, b / 5


def generate_ccg_palette(curses_palette: List[int],
                         reference_y: float) -> Palette:
    cbcr_palette: Dict[float, Tuple[float, int]] = {}
    for curses_color in curses_palette:
        r, g, b = ncurses_color_to_rgb(curses_color)
        # drop grayscale
        if r == g == b:
            continue
        h, _, y = hsluv.rgb_to_hsluv((r, g, b))
        # this is to keep the code compatible with earlier versions of XEP-0392
        y = y / 100
        key = round(h)
        try:
            existing_y, *_ = cbcr_palette[key]
        except KeyError:
            pass
        else:
            if abs(existing_y - reference_y) <= abs(y - reference_y):
                continue
        cbcr_palette[key] = y, curses_color
    return {
        angle: curses_color
        for angle, (_, curses_color) in cbcr_palette.items()
    }


def text_to_angle(text: str) -> float:
    hf = hashlib.sha1()
    hf.update(text.encode("utf-8"))
    hue = int.from_bytes(hf.digest()[:2], "little")
    return hue / 65535 * 360


def ccg_palette_lookup(palette: Palette, angle: float) -> int:
    # try quick lookup first
    try:
        return palette[round(angle)]
    except KeyError:
        pass

    best_metric = float("inf")
    best = None
    for anglep, color in palette.items():
        metric = abs(anglep - angle)
        if metric < best_metric:
            best_metric = metric
            best = color

    if best is None:
        raise ColorNotFoundInPalette()

    return best


def ccg_text_to_color(palette: Palette | None, text: str) -> int:
    if palette is None:
        raise ColorNotFoundInPalette()
    angle = text_to_angle(text)
    return ccg_palette_lookup(palette, angle)



"""
This plugin allows you to ping an entity.

Command
-------

.. glossary::

    /ping
        **Usage (globally):** ``/ping <jid>``

        **Usage (in a MUC tab):** ``/ping <jid or nick>``

        **Usage (in a conversation tab):** ``/ping [jid]``

        Globally, you can do ``/ping jid@example.com`` to get a ping.

        In a MUC, you can either do it to a JID or a nick (``/ping nick`` or ``/ping
        jid@example.com``).

        In a private or a direct conversation, you can do ``/ping`` to ping
        the current interlocutor.
"""

from slixmpp import InvalidJID, JID
from slixmpp.exceptions import IqTimeout
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.windows.inputs import Input
from poezio.roster import roster
from poezio.contact import Contact, Resource
from poezio.core.structs import Completion
from poezio import tabs
import time


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'ping',
            self.command_ping,
            usage='<jid>',
            help='Send an XMPP ping to jid (see XEP-0199).',
            short='Send a ping',
            completion=self.completion_ping)
        self.api.add_tab_command(
            tabs.MucTab,
            'ping',
            self.command_muc_ping,
            usage='<jid|nick>',
            help='Send an XMPP ping to jid or nick (see XEP-0199).',
            short='Send a ping.',
            completion=self.completion_muc_ping)
        self.api.add_tab_command(
            tabs.RosterInfoTab,
            'ping',
            self.command_roster_ping,
            usage='<jid>',
            help='Send an XMPP ping to jid (see XEP-0199).',
            short='Send a ping.',
            completion=self.completion_ping)
        for _class in (tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                _class,
                'ping',
                self.command_private_ping,
                usage='[jid]',
                help='Send an XMPP ping to the current interlocutor or the given JID.',
                short='Send a ping',
                completion=self.completion_ping)

    @command_args_parser.raw
    async def command_ping(self, arg: str) -> None:
        if not arg:
            return self.core.command.help('ping')
        try:
            jid = JID(arg)
        except InvalidJID:
            self.api.information('Invalid JID: %s' % arg, 'Error')
            return
        start = time.time()

        try:
            iq = await self.core.xmpp.plugin['xep_0199'].send_ping(
                jid=jid, timeout=10
            )
            delay = time.time() - start
            error = False
            reply = ''
            if iq['type'] == 'error':
                error_condition = iq['error']['condition']
                reply = error_condition
                # These IQ errors are not ping errors:
                # 'service-unavailable': official "not supported" response as of RFC6120 (§8.4) and XEP-0199 (§4.1)
                # 'feature-not-implemented': inoffcial not-supported response from many clients
                if error_condition not in ('service-unavailable',
                                           'feature-not-implemented'):
                    error = True
                error_text = iq['error']['text']
                if error_text:
                    reply = '%s: %s' % (error_condition, error_text)
            if error:
                message = '%s did not respond to ping: %s' % (jid, reply)
            else:
                reply = ' (%s)' % reply if reply else ''
                message = '%s responded to ping after %ss%s' % (
                    jid, round(delay, 4), reply)
            self.api.information(message, 'Info')
        except IqTimeout:
            self.api.information(
                '%s did not respond to ping after 10s: timeout' % jid,
                'Info'
            )

    def completion_muc_ping(self, the_input: Input) -> Completion | None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.MucTab):
            return None
        users = [user.nick for user in tab.users]
        users.extend(self.resources())
        return Completion(the_input.auto_completion, users, '', quotify=False)

    @command_args_parser.raw
    def command_private_ping(self, arg: str) -> None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.OneToOneTab):
            return
        jid = arg
        if not arg:
            jid = str(tab.jid)
        self.core.register_task(
            self.command_ping(jid)
        )

    @command_args_parser.raw
    def command_muc_ping(self, arg: str) -> None:
        if not arg:
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.MucTab):
            return
        user = tab.get_user_by_name(arg)
        if user:
            jid = tab.jid
            jid.resource = user.nick
        else:
            try:
                jid = JID(arg)
            except InvalidJID:
                self.api.information('Invalid JID: %s' % arg, 'Error')
                return
        self.core.register_task(
            self.command_ping(jid.full)
        )

    @command_args_parser.raw
    def command_roster_ping(self, arg: str) -> None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.RosterInfoTab):
            return
        jid = None
        if arg:
            jid = arg
        else:
            current = tab.selected_row
            if isinstance(current, Resource):
                jid = current.jid
            elif isinstance(current, Contact):
                res = current.get_highest_priority_resource()
                if res is not None:
                    jid = res.jid
        if jid:
            self.core.register_task(
                self.command_ping(jid)
            )

    def resources(self) -> list[str]:
        resources = []
        for contact in roster.get_contacts():
            for resource in contact.resources:
                resources.append(str(resource.jid))
        return resources

    def completion_ping(self, the_input: Input) -> Completion:
        return Completion(
            the_input.auto_completion, self.resources(), '', quotify=False)



"""
Tabs management module

Provide a class holding the current tabs of the application.
Supported list modification operations:
    - Appending a tab
    - Deleting a tab (and going back to the previous one)
    - Inserting a tab from a position to another
    - Replacing the whole tab list with another (used for rearranging the
      list from outside)

This class holds a cursor to the current tab, which allows:
    - Going left (prev()) or right (next()) in the list, cycling
    - Getting a reference to the current tab
    - Setting the current tab by index or reference

It supports the poezio "gap tab" concept, aka empty tabs taking a space in the
tab list in order to avoid shifting the tab numbers when closing a tab.
Example tab list: [0|1|2|3|4]
We then close the tab 3: [0|1|2|4]
The tab has been closed and replaced with a "gap tab", which the user cannot
switch to, but which avoids shifting numbers (in the case above, the list would
have become [0|1|2|3], with the tab "4" renumbered to "3" if gap tabs are
disabled.
"""

from typing import Type, Optional, Union, TypeVar, cast, overload
from collections import defaultdict
from collections.abc import Iterator
from slixmpp import JID
from poezio import tabs
from poezio.events import EventHandler

T = TypeVar('T', bound=tabs.Tab)


class Tabs:
    """
    Tab list class
    """
    __slots__ = [
        '_current_index',
        '_current_tab',
        '_tabs',
        '_tab_jids',
        '_tab_types',
        '_tab_names',
        '_previous_tab',
        '_events',
    ]

    def __init__(self, events: EventHandler, initial_tab: tabs.Tab) -> None:
        """
        Initialize the Tab list. Even though the list is initially
        empty, all methods are only valid once append() has been called
        once. Otherwise, mayhem is expected.
        """
        # cursor
        self._current_index: int = 0
        self._current_tab: tabs.Tab = initial_tab

        self._previous_tab: Optional[tabs.Tab] = None
        self._tabs: list[tabs.Tab] = []
        self._tab_jids: dict[JID, tabs.Tab] = dict()
        self._tab_types: dict[Type[tabs.Tab], list[tabs.Tab]] = defaultdict(list)
        self._tab_names: dict[str, tabs.Tab] = dict()
        self._events: EventHandler = events

    def __len__(self) -> int:
        return len(self._tabs)

    def __iter__(self) -> Iterator[tabs.Tab]:
        return iter(self._tabs)

    @overload
    def __getitem__(self, index: int | str) -> tabs.Tab | None:
        ...

    @overload
    def __getitem__(self, index: slice) -> list[tabs.Tab]:
        ...

    def __getitem__(self, index: int | str | slice) -> tabs.Tab | None | list[tabs.Tab]:
        if isinstance(index, str):
            return self.by_name(index)
        return self._tabs[index]

    def first(self) -> tabs.Tab:
        """Return the Roster tab"""
        return self._tabs[0]

    @property
    def current_index(self) -> int:
        """Current tab index"""
        return self._current_index

    def set_current_index(self, value: int) -> bool:
        """Set the current tab index"""
        if 0 <= value < len(self._tabs):
            tab = self._tabs[value]
            return self.set_current_tab(tab)
        return False

    @property
    def current_tab(self) -> tabs.Tab:
        """Current tab"""
        return self._current_tab

    def set_current_tab(self, tab: tabs.Tab) -> bool:
        """Set the current tab"""
        if (not isinstance(tab, tabs.GapTab)
                and 0 <= tab.nb < len(self._tabs)):
            self._store_previous()
            self._current_index = tab.nb
            self._current_tab = tab
            self._events.trigger(
                'tab_change',
                old_tab=self._previous_tab,
                new_tab=self._current_tab)
            return True
        return False

    def get_tabs(self) -> list[tabs.Tab]:
        """Return the tab list"""
        return self._tabs

    def by_jid(self, jid: JID) -> tabs.Tab | None:
        """Get a tab with a specific jid"""
        return self._tab_jids.get(jid)

    def by_name(self, name: str) -> tabs.Tab | None:
        """Get a tab with a specific name"""
        return self._tab_names.get(name)

    def by_class(self, cls: Type[T]) -> list[T]:
        """Get all the tabs of a class"""
        return cast(list[T], list(self._tab_types.get(cls, [])))

    def find_match(self, name: str) -> tabs.Tab | None:
        """Get a tab using extended matching (tab.matching_name())"""

        def transform(tab_index: int) -> int:
            """Wrap the value of the range around the current index"""
            return (tab_index + self._current_index + 1) % len(self._tabs)

        for i in map(transform, range(len(self._tabs) - 1)):
            for tab_name in self._tabs[i].matching_names():
                if tab_name[1] and name in tab_name[1].lower():
                    return self._tabs[i]
        return None

    def find_by_unique_prefix(self, prefix: str) -> tuple[bool, Optional[tabs.Tab]]:
        """
        Get a tab by its unique name prefix, ignoring case.

        :return: A tuple indicating the presence of any match, as well as the
            uniquely matched tab (if any).

        The first element, a boolean, in the returned tuple indicates whether
        at least one tab matched.

        The second element (a Tab) in the returned tuple is the uniquely
        matched tab, if any. If multiple or no tabs match the prefix, the
        second element in the tuple is :data:`None`.
        """

        # TODO: should this maybe use something smarter than .lower()?
        # something something stringprep?
        prefix = prefix.lower()
        candidate = None
        any_matched = False
        for tab in self._tabs:
            if not tab.name.lower().startswith(prefix):
                continue
            any_matched = True
            if candidate is not None:
                # multiple tabs match -> return None
                return True, None
            candidate = tab

        return any_matched, candidate

    def by_name_and_class(self, name: Union[str, JID],
                          cls: Type[T]) -> Optional[T]:
        """Get a tab with its name and class"""
        if isinstance(name, JID):
            str_name = name.full
        else:
            str_name = name
        str
        cls_tabs = self._tab_types.get(cls, [])
        for tab in cls_tabs:
            if tab.name == str_name:
                return cast(T, tab)
        return None

    def _rebuild(self) -> None:
        self._tab_jids = dict()
        self._tab_types = defaultdict(list)
        self._tab_names = dict()
        for tab in self._tabs:
            for cls in _get_tab_types(tab):
                self._tab_types[cls].append(tab)
            if hasattr(tab, 'jid'):
                self._tab_jids[tab.jid] = tab
            self._tab_names[tab.name] = tab
        self._update_numbers()

    def replace_tabs(self, new_tabs: list[tabs.Tab]) -> bool:
        """
        Replace the current tab list with another, and
        rebuild the mappings.
        """
        if self._current_tab not in new_tabs:
            return False
        self._tabs = new_tabs
        self._rebuild()
        return True

    def _inc_cursor(self) -> None:
        self._current_index = (self._current_index + 1) % len(self._tabs)
        self._current_tab = self._tabs[self._current_index]

    def _dec_cursor(self) -> None:
        self._current_index = (self._current_index - 1) % len(self._tabs)
        self._current_tab = self._tabs[self._current_index]

    def _store_previous(self) -> None:
        self._previous_tab = self._current_tab

    def next(self) -> None:
        """Go to the right of the tab list (circular)"""
        self._store_previous()
        self._inc_cursor()
        while isinstance(self.current_tab, tabs.GapTab):
            self._inc_cursor()
        self._events.trigger(
            'tab_change',
            old_tab=self._previous_tab,
            new_tab=self._current_tab)

    def prev(self) -> None:
        """Go to the left of the tab list (circular)"""
        self._store_previous()
        self._dec_cursor()
        while isinstance(self.current_tab, tabs.GapTab):
            self._dec_cursor()
        self._events.trigger(
            'tab_change',
            old_tab=self._previous_tab,
            new_tab=self._current_tab)

    def append(self, tab: tabs.Tab) -> None:
        """
        Add a tab to the list
        """
        if not self._tabs:
            tab.nb = 0
            self._current_tab = tab
        else:
            tab.nb = self._tabs[-1].nb + 1
        self._tabs.append(tab)
        for cls in _get_tab_types(tab):
            self._tab_types[cls].append(tab)
        if isinstance(tab, tabs.ChatTab):
            self._tab_jids[tab.jid] = tab
        self._tab_names[tab.name] = tab

    def delete(self, tab: tabs.Tab, gap: bool = False) -> None:
        """Remove a tab"""
        if isinstance(tab, tabs.RosterInfoTab):
            return

        if gap:
            self._tabs[tab.nb] = tabs.GapTab()
        else:
            self._tabs.remove(tab)

        is_current = tab is self.current_tab

        for cls in _get_tab_types(tab):
            self._tab_types[cls].remove(tab)
        if hasattr(tab, 'jid'):
            del self._tab_jids[tab.jid]
        del self._tab_names[tab.name]

        if gap:
            self._collect_trailing_gaptabs()
        else:
            self._update_numbers()

        if tab is self._previous_tab:
            self._previous_tab = None
        if is_current:
            self.restore_previous_tab()
            self._previous_tab = None
        self._validate_current_index()

    def restore_previous_tab(self) -> None:
        """Restore the previous tab"""

        if self._previous_tab:
            if not self.set_current_tab(self._previous_tab):
                self.set_current_index(0)
        else:
            self.set_current_index(0)

    def _validate_current_index(self) -> None:
        if not 0 <= self._current_index < len(
                self._tabs) or not self.current_tab:
            self.prev()  # pylint: disable=not-callable

    def _collect_trailing_gaptabs(self) -> None:
        """Remove trailing gap tabs if any"""
        i = len(self._tabs) - 1
        while isinstance(self._tabs[i], tabs.GapTab):
            self._tabs.pop()
            i -= 1

    def _update_numbers(self) -> None:
        for i, tab in enumerate(self._tabs):
            tab.nb = i
        self._current_index = self._current_tab.nb

    # Moving tabs around #

    def update_gaps(self, enable_gaps: bool) -> None:
        """
        Remove the present gaps from the list if enable_gaps is False.
        """
        if not enable_gaps:
            self._tabs = [tab for tab in self._tabs if tab]
            self._update_numbers()

    def _insert_nogaps(self, old_pos: int, new_pos: int) -> bool:
        """
        Move tabs without creating gaps
        old_pos: old position of the tab
        new_pos: desired position of the tab
        """
        tab = self._tabs[old_pos]
        if new_pos < old_pos:
            self._tabs.pop(old_pos)
            self._tabs.insert(new_pos, tab)
        elif new_pos > old_pos:
            self._tabs.insert(new_pos, tab)
            self._tabs.remove(tab)
        else:
            return False
        return True

    def _insert_gaps(self, old_pos: int, new_pos: int) -> bool:
        """
        Move tabs and create gaps in the eventual remaining space
        old_pos: old position of the tab
        new_pos: desired position of the tab
        """
        tab = self._tabs[old_pos]
        target = None if new_pos >= len(self._tabs) else self._tabs[new_pos]
        if not target:
            if new_pos < len(self._tabs):
                old_tab = self._tabs[old_pos]
                self._tabs[new_pos], self._tabs[
                    old_pos] = old_tab, tabs.GapTab()
            else:
                self._tabs.append(self._tabs[old_pos])
                self._tabs[old_pos] = tabs.GapTab()
        else:
            if new_pos > old_pos:
                self._tabs.insert(new_pos, tab)
                self._tabs[old_pos] = tabs.GapTab()
            elif new_pos < old_pos:
                self._tabs[old_pos] = tabs.GapTab()
                self._tabs.insert(new_pos, tab)
            else:
                return False
            i = self._tabs.index(tab)
            done = False
            # Remove the first Gap on the right in the list
            # in order to prevent global shifts when there is empty space
            while not done:
                i += 1
                if i >= len(self._tabs):
                    done = True
                elif not self._tabs[i]:
                    self._tabs.pop(i)
                    done = True
        self._collect_trailing_gaptabs()
        return True

    def insert_tab(self, old_pos: int, new_pos: int = 99999, gaps: bool = False) -> bool:
        """
        Insert a tab at a position, changing the number of the following tabs
        returns False if it could not move the tab, True otherwise
        """
        if (old_pos <= 0 or old_pos >= len(self._tabs) or new_pos <= 0
                or new_pos == old_pos or not self._tabs[old_pos]):
            return False
        if gaps:
            result = self._insert_gaps(old_pos, new_pos)
        else:
            result = self._insert_nogaps(old_pos, new_pos)
        self._update_numbers()
        return result


def _get_tab_types(tab: tabs.Tab) -> list[Type[tabs.Tab]]:
    """Return all parent classes of a tab type"""
    types = []
    current_cls = tab.__class__
    while current_cls != tabs.Tab:
        types.append(current_cls)
        current_cls = current_cls.__bases__[0]
    return types
